@@ -542,7 +542,7 @@ mark_subst_opts(cmd)
   cmd->global = FALSE;
   cmd->print = FALSE;
   cmd->eval = FALSE;
-  cmd->numb = FALSE;
+  cmd->numb = 0;
   cmd->fp = NULL;
 
   for (;;)
