@@ -423,7 +423,7 @@ release_append_queue()
   for (p=append_head; p; p=q)
     {
       if (p->free)
-        FREE(p->text);
+	FREE(p->text);
 
       q = p->next;
       FREE(p);
@@ -540,9 +540,9 @@ open_next_file(name, input)
 
 #ifdef HAVE_FCHMOD
       {
-        struct stat st;
-        fstat (fileno (input->fp), &st);
-        fchmod (output_fd, st.st_mode);
+	struct stat st;
+	fstat (fileno (input->fp), &st);
+	fchmod (output_fd, st.st_mode);
       }
 #endif
 
@@ -716,7 +716,7 @@ match_an_address_p(addr, input)
     case addr_is_step_mod:
       /* reminder: these are only meaningful for a2 addresses */
       /* a2->addr_number needs to be recomputed each time a1 address
-         matches for the step and step_mod types */
+	 matches for the step and step_mod types */
       return (addr->addr_number <= input->line_number);
 
     case addr_is_last:
@@ -830,14 +830,61 @@ do_list(line_len)
   ck_fwrite("$\n", 1, 2, output_file);
 }
 
+static enum replacement_types append_replacement P_((struct line *, struct replacement *,
+				                     struct re_registers *,
+				                     enum replacement_types));
+static enum replacement_types
+append_replacement (buf, p, regs, repl_mod)
+  struct line *buf;
+  struct replacement *p;
+  struct re_registers *regs;
+  enum replacement_types repl_mod;
+{
+  for (; p; p=p->next)
+    {
+      int i = p->subst_id;
+      enum replacement_types curr_type;
+
+      /* Apply a \[lu] modifier that was given earlier, but which we
+	 have not had yet the occasion to apply.  But don't do it
+	 if this replacement has a modifier of its own. */
+      curr_type = (p->repl_type & repl_modifiers)
+	? p->repl_type
+	: p->repl_type | repl_mod;
+
+      repl_mod = 0;
+      if (p->prefix_length)
+	{
+	  str_append_modified(buf, p->prefix, p->prefix_length,
+			      curr_type);
+	  curr_type &= ~repl_modifiers;
+	}
+
+      if (0 <= i)
+	if (regs->end[i] == regs->start[i] && p->repl_type & repl_modifiers)
+	  /* Save this modifier, we shall apply it later.
+	     e.g. in s/()([a-z])/\u\1\2/
+	     the \u modifier is applied to \2, not \1 */
+	  repl_mod = curr_type & repl_modifiers;
+
+	else
+	  str_append_modified(buf, line.active + regs->start[i],
+			      CAST(size_t)(regs->end[i] - regs->start[i]),
+			      curr_type);
+    }
+
+  return repl_mod;
+}
+
 static void do_subst P_((struct subst *));
 static void
 do_subst(sub)
   struct subst *sub;
 {
-  size_t start = 0;	/* where to start scan for (next) match in LINE */
+  size_t start = 0;     /* where to start scan for (next) match in LINE */
   size_t last_end = 0;  /* where did the last successful match end in LINE */
-  countT count = 0;	/* number of matches found */
+  countT count = 0;     /* number of matches found */
+  flagT again = TRUE;
 
 #define MAX_BACKREFERENCES 10
   static struct re_registers regs;
@@ -851,7 +898,7 @@ do_subst(sub)
   if (!match_regex(sub->regx, line.active, line.length, start,
 		   &regs, MAX_BACKREFERENCES))
     return;
-  
+
   if (!sub->replacement && sub->numb <= 1)
     if (regs.start[0] == 0 && !sub->global)
       {
@@ -861,7 +908,7 @@ do_subst(sub)
 	line.active += regs.end[0];
 	line.length -= regs.end[0];
 	line.alloc -= regs.end[0];
-	goto did_subst;
+	goto post_subst;
       }
     else if (regs.end[0] == line.length)
       {
@@ -869,12 +916,11 @@ do_subst(sub)
 	replaced = TRUE;
 
 	line.length = regs.start[0];
-	goto did_subst;
+	goto post_subst;
       }
 
   do
     {
-      struct replacement *p;
       enum replacement_types repl_mod = 0;
 
       size_t offset = regs.start[0];
@@ -885,82 +931,47 @@ do_subst(sub)
 	str_append(&s_accum, line.active + start, offset - start);
 
       /* If we're counting up to the Nth match, are we there yet?
-         And even if we are there, there is another case we have to
+	 And even if we are there, there is another case we have to
 	 skip: are we matching an empty string immediately following
-         another match?
+	 another match?
 
-         This latter case avoids that baaaac, when passed through
-         s,a*,x,g, gives `xbxxcx' instead of xbxcx.  This behavior is
-         unacceptable because it is not consistently applied (for
-         example, `baaaa' gives `xbx', not `xbxx'). */
-      if ((count > 0 && offset == last_end && matched == 0)
-	  || ++count < sub->numb)
+	 This latter case avoids that baaaac, when passed through
+	 s,a*,x,g, gives `xbxxcx' instead of xbxcx.  This behavior is
+	 unacceptable because it is not consistently applied (for
+	 example, `baaaa' gives `xbx', not `xbxx'). */
+      if ((matched > 0 || count == 0 || offset > last_end)
+	  && (last_end = regs.end[0], ++count >= sub->numb))
 	{
-	  /* If the match was vacuous, skip ahead one character
-	   * anyway.
-	   */
-	  if (matched == 0 && offset < line.length)
-	    matched = 1;
+	  /* We found a match, set the `replaced' flag. */
+	  replaced = TRUE;
 
-	  str_append(&s_accum, line.active + offset, matched);
-	  start = offset + matched;
-	  continue;
+	  /* Now expand the replacement string into the output string. */
+	  repl_mod = append_replacement (&s_accum, sub->replacement, &regs, repl_mod);
+	  again = sub->global;
 	}
-
-      /* We found a match, set the `replaced' flag. */
-      replaced = TRUE;
-
-      last_end = regs.end[0];
-
-      /* Now expand the replacement string into the output string. */
-      for (p=sub->replacement; p; p=p->next)
+      else
 	{
-	  int i = p->subst_id;
-	  enum replacement_types curr_type;
-
-	  /* Apply a \[lu] modifier that was given earlier, but which we
-	     have not had yet the occasion to apply.  But don't do it
-	     if this replacement has a modifier of its own. */
-	  curr_type = (p->repl_type & repl_modifiers)
-	    ? p->repl_type
-	    : p->repl_type | repl_mod;
-
-	  repl_mod = 0;
-	  if (p->prefix_length)
+	  /* The match was not replaced.  Copy the text until its
+	     end; if it was vacuous, skip over one character and
+	     add that character to the output.  */
+	  if (matched == 0)
 	    {
-	      str_append_modified(&s_accum, p->prefix, p->prefix_length,
-				  curr_type);
-	      curr_type &= ~repl_modifiers;
+	      if (start < line.length)
+		matched = 1;
+	      else
+		break;
 	    }
 
-	  if (0 <= i)
-	    if (regs.end[i] == regs.start[i] && p->repl_type & repl_modifiers)
-	      /* Save this modifier, we shall apply it later.
-		 e.g. in s/()([a-z])/\u\1\2/
-		 the \u modifier is applied to \2, not \1 */
-	      repl_mod = curr_type & repl_modifiers;
-
-	    else
-	      str_append_modified(&s_accum, line.active + regs.start[i],
-				  CAST(size_t)(regs.end[i] - regs.start[i]),
-				  curr_type);
+	  str_append(&s_accum, line.active + offset, matched);
 	}
 
-      start = regs.end[0];
-      if (!sub->global || start == line.length)
-	break;
-
-      /* If the match was vacuous, skip over one character
-       * and add that character to the output.
-       */
-      if (regs.start[0] == regs.end[0])
-	{
-	  str_append(&s_accum, line.active + offset, 1);
-	  ++start;
-	}
+      /* Start after the match.  */
+      start = offset + matched;
     }
-  while (match_regex(sub->regx, line.active, line.length, start,
-		     &regs, MAX_BACKREFERENCES));
+  while (again
+	 && start <= line.length
+	 && match_regex(sub->regx, line.active, line.length, start,
+			&regs, MAX_BACKREFERENCES));
 
   /* Copy stuff to the right of the last match into the output string. */
   if (start < line.length)
@@ -970,18 +981,21 @@ do_subst(sub)
   /* Exchange line and s_accum.  This can be much cheaper
      than copying s_accum.active into line.text (for huge lines). */
   line_exchange(&line, &s_accum);
- did_subst:
-  
+
   /* Finish up. */
+  if (count < sub->numb)
+    return;
+
+ post_subst:
   if (sub->print & 1)
     output_line(line.active, line.length, line.chomped, output_file);
-  
-  if (sub->eval) 
+
+  if (sub->eval)
     {
 #ifdef HAVE_POPEN
       FILE *pipe;
       s_accum.length = 0;
-      
+
       str_append (&line, "", 1);
       pipe = popen(line.active, "r");
       
@@ -1317,11 +1331,11 @@ execute_program(vec, input)
 
 	    case 'W':
 	      if (cur_cmd->x.fp)
-	        {
+		{
 		  char *p = memchr(line.active, '\n', line.length);
 		  output_line(line.active, p ? p - line.active : line.length,
 			      p ? 1 : line.chomped, cur_cmd->x.fp);
-	        }
+		}
 	      break;
 
 	    case 'x':
@@ -1331,81 +1345,81 @@ execute_program(vec, input)
 	    case 'y':
 	      {
 #if defined MBS_SUPPORT && !defined REG_PERL
-               if (MB_CUR_MAX > 1)
-                 {
-                   int idx, prev_idx; /* index in the input line.  */
-                   char **trans;
-                   mbstate_t cur_stat;
-                   memset(&cur_stat, 0, sizeof(mbstate_t));
-                   for (idx = 0; idx < line.length;)
-                     {
-                       int mbclen, i;
-                       mbclen = mbrlen(line.active + idx, line.length - idx,
-                                       &cur_stat);
-                       /* An invalid sequence, or a truncated multibyte
-                          character.  We treat it as a singlebyte character.
-                       */
-                       if (mbclen == (size_t) -1 || mbclen == (size_t) -2
-                           || mbclen == 0)
-                         mbclen = 1;
+	       if (MB_CUR_MAX > 1)
+		 {
+		   int idx, prev_idx; /* index in the input line.  */
+		   char **trans;
+		   mbstate_t cur_stat;
+		   memset(&cur_stat, 0, sizeof(mbstate_t));
+		   for (idx = 0; idx < line.length;)
+		     {
+		       int mbclen, i;
+		       mbclen = mbrlen(line.active + idx, line.length - idx,
+				       &cur_stat);
+		       /* An invalid sequence, or a truncated multibyte
+			  character.  We treat it as a singlebyte character.
+		       */
+		       if (mbclen == (size_t) -1 || mbclen == (size_t) -2
+			   || mbclen == 0)
+			 mbclen = 1;
 
-                       trans = cur_cmd->x.translatemb;
-                       /* `i' indicate i-th translate pair.  */
-                       for (i = 0; trans[2*i] != NULL; i++)
-                         {
-                           if (strncmp(line.active + idx, trans[2*i], mbclen)
-                               == 0)
-                             {
-                               flagT move_remain_buffer = FALSE;
-                               int trans_len = strlen(trans[2*i+1]);
+		       trans = cur_cmd->x.translatemb;
+		       /* `i' indicate i-th translate pair.  */
+		       for (i = 0; trans[2*i] != NULL; i++)
+			 {
+			   if (strncmp(line.active + idx, trans[2*i], mbclen)
+			       == 0)
+			     {
+			       flagT move_remain_buffer = FALSE;
+			       int trans_len = strlen(trans[2*i+1]);
 
-                               if (mbclen < trans_len)
-                                 {
-                                   int new_len;
-                                   new_len = line.length + 1 + trans_len - mbclen;
-                                   /* We must extend the line buffer.  */
-                                   if (line.alloc < new_len)
-                                     {
-                                       /* And we must resize the buffer.  */
-                                       resize_line(&line, new_len);
-                                     }
-                                   move_remain_buffer = TRUE;
-                                 }
-                               else if (mbclen > trans_len)
-                                 {
-                                   /* We must truncate the line buffer.  */
-                                   move_remain_buffer = TRUE;
-                                 }
-                               prev_idx = idx;
-                               if (move_remain_buffer)
-                                 {
-                                   int move_len, move_offset;
-                                   char *move_from, *move_to;
-                                   /* Move the remaining with \0.  */
-                                   move_from = line.active + idx + mbclen;
-                                   move_to = line.active + idx + trans_len;
-                                   move_len = line.length + 1 - idx - mbclen;
-                                   move_offset = trans_len - mbclen;
-                                   memmove(move_to, move_from, move_len);
-                                   line.length += move_offset;
-                                   idx += move_offset;
-                                 }
-                               strncpy(line.active + prev_idx, trans[2*i+1],
-                                       trans_len);
-                               break;
-                             }
-                         }
-                       idx += mbclen;
-                     }
-                 }
-               else
+			       if (mbclen < trans_len)
+				 {
+				   int new_len;
+				   new_len = line.length + 1 + trans_len - mbclen;
+				   /* We must extend the line buffer.  */
+				   if (line.alloc < new_len)
+				     {
+				       /* And we must resize the buffer.  */
+				       resize_line(&line, new_len);
+				     }
+				   move_remain_buffer = TRUE;
+				 }
+			       else if (mbclen > trans_len)
+				 {
+				   /* We must truncate the line buffer.  */
+				   move_remain_buffer = TRUE;
+				 }
+			       prev_idx = idx;
+			       if (move_remain_buffer)
+				 {
+				   int move_len, move_offset;
+				   char *move_from, *move_to;
+				   /* Move the remaining with \0.  */
+				   move_from = line.active + idx + mbclen;
+				   move_to = line.active + idx + trans_len;
+				   move_len = line.length + 1 - idx - mbclen;
+				   move_offset = trans_len - mbclen;
+				   memmove(move_to, move_from, move_len);
+				   line.length += move_offset;
+				   idx += move_offset;
+				 }
+			       strncpy(line.active + prev_idx, trans[2*i+1],
+				       trans_len);
+			       break;
+			     }
+			 }
+		       idx += mbclen;
+		     }
+		 }
+	       else
 #endif /* MBS_SUPPORT */
-                 {
-                   unsigned char *p, *e;
-                   p = CAST(unsigned char *)line.active;
-                   for (e=p+line.length; p<e; ++p)
-                     *p = cur_cmd->x.translate[*p];
-                 }
+		 {
+		   unsigned char *p, *e;
+		   p = CAST(unsigned char *)line.active;
+		   for (e=p+line.length; p<e; ++p)
+		     *p = cur_cmd->x.translate[*p];
+		 }
 	      }
 	      break;
 
